/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.bonita.connector.core;

import java.io.File;
import java.lang.reflect.Modifier;
import java.net.URI;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Enumeration;
import java.util.List;
import java.util.Locale;
import java.util.ResourceBundle;
import java.util.jar.JarEntry;
import java.util.jar.JarFile;

/**
 * 
 * @author chaffotm
 *
 */
public class ConnectorAPI {

  private List<ConnectorDescription> connectors;
  private Locale currentLocale;

  public ConnectorAPI(JarFile jarFile) throws Exception {
  	this(jarFile, Locale.ENGLISH);
  }

  public ConnectorAPI(JarFile jarFile, Locale locale) throws Exception {
    this.currentLocale = locale;
    this.connectors = new ArrayList<ConnectorDescription>();
  	loadConnectors(jarFile);
  }

  public ConnectorAPI(URI uri) throws Exception {
  	this(uri, Locale.ENGLISH);
  }

  public ConnectorAPI(URI uri, Locale locale) throws Exception {
    this.currentLocale = locale;
    this.connectors = new ArrayList<ConnectorDescription>();
  	loadConnectors(uri);
  }

  public ConnectorAPI(String jarPath) throws Exception {
  	this(jarPath, Locale.ENGLISH);
  }

  public ConnectorAPI(String jarPath, Locale locale) throws Exception {
    this.currentLocale = locale;
    this.connectors = new ArrayList<ConnectorDescription>();
  	loadConnectors(jarPath);
  }

  protected ConnectorAPI(List<Class<?>> classes) throws Exception {
    this(classes, Locale.ENGLISH);
  }

  @SuppressWarnings("unchecked")
  protected ConnectorAPI(List<Class<?>> classes, Locale locale) throws Exception {
    this.currentLocale = locale;
    this.connectors = new ArrayList<ConnectorDescription>();
    List<ConnectorDescription> tempList = new ArrayList<ConnectorDescription>();
    for (Class<?> clazz : classes) {
      if (isAConnector(clazz)) {
        ConnectorDescription temp =
          new ConnectorDescription((Class<? extends Connector>)clazz, currentLocale);
        String id = temp.getId();
        if (idExists(id)) {
          throw new IllegalArgumentException("Cannot load " + clazz.getName()
              + " because a similar id exists");
        }
        tempList.add(temp);
      }
    }
    connectors.addAll(tempList);
  }

  @SuppressWarnings("unchecked")
  public void loadConnectors(JarFile jarFile) throws Exception {
    Enumeration<JarEntry> files = jarFile.entries();
    List<ConnectorDescription> tempList = new ArrayList<ConnectorDescription>();
    while (files.hasMoreElements()) {
      JarEntry file = files.nextElement();
      if (!file.isDirectory()) {
        String fileName = file.getName();
        if (fileName.endsWith(".class")) {
          fileName = fileName.replace('/', '.');
          int length = fileName.length();
          String className = fileName.substring(0, length - 6);
          if (className.startsWith("org.ow2.bonita.connector")
          		&& !className.contains("$")) {
            Class<?> clazz = Class.forName(className);
            if (isAConnector(clazz)) {
              ConnectorDescription temp =
                new ConnectorDescription((Class<? extends Connector>)clazz, currentLocale);
              String id = temp.getId();
              if (idExists(id)) {
                throw new IllegalArgumentException("Cannot load " + clazz.getName()
                    + " because a similar id exists");
              }
              tempList.add(temp);
            }
          }
        }
      }
    }
    connectors.addAll(tempList);
  }

  private boolean idExists(String id) {
    boolean exists = false;
    for (ConnectorDescription connector : connectors) {
      if (id.equals(connector.getId())) {
        exists = true;
        break;
      }
    }
    return exists;
  }

  public void loadConnectors(String jarPath) throws Exception {
    loadConnectors(new JarFile(jarPath));
  }

  public void loadConnectors(URI uri) throws Exception {
    File file = new File(uri);
    loadConnectors(new JarFile(file));
  }

  public List<ConnectorDescription> getAllConnectors() {
    return connectors;
  }

  public List<ConnectorDescription> getJavaConnectors() {
  	List<ConnectorDescription> javaconnectors = new ArrayList<ConnectorDescription>();
  	for (ConnectorDescription connector : connectors) {
  		Class<? extends Connector> clazz = connector.getConnectorClass();
  		if (isConnector(clazz) && !isSimpleMapper(clazz)
  				&& !isProcessConnector(clazz)) {
  			javaconnectors.add(connector);
  		}
    }
  	return javaconnectors;
  }

  public List<ConnectorDescription> getRoleMappers() {
  	List<ConnectorDescription> rolemappers = new ArrayList<ConnectorDescription>();
  	for (ConnectorDescription connector : connectors) {
  		Class<? extends Connector> clazz = connector.getConnectorClass();
  		if (isRoleMapper(clazz)) {
  			rolemappers.add(connector);
  		}
    }
  	return rolemappers;
  }

  public List<ConnectorDescription> getMappers() {
  	List<ConnectorDescription> rolemappers = new ArrayList<ConnectorDescription>();
  	for (ConnectorDescription connector : connectors) {
  		Class<? extends Connector> clazz = connector.getConnectorClass();
  		if (isMapper(clazz)) {
  			rolemappers.add(connector);
  		}
    }
  	return rolemappers;
  }

  private boolean isConnector(Class<?> clazz) {
  	try {
  		Object connector = clazz.newInstance();
  		return connector instanceof Connector;
  	} catch (Exception e) {
			return false;
		}
  }

  private boolean isSimpleMapper(Class<?> clazz) {
  	try {
  		Object connector = clazz.newInstance();
  		return connector instanceof SimpleMapper;
  	} catch (Exception e) {
			return false;
		}
  }

  private boolean isMapper(Class<?> clazz) {
  	try {
  		Object connector = clazz.newInstance();
  		return connector instanceof Mapper;
  	} catch (Exception e) {
			return false;
		}
  }

  private boolean isRoleMapper(Class<?> clazz) {
  	try {
  		Object connector = clazz.newInstance();
  		return connector instanceof RoleMapperConnector;
  	} catch (Exception e) {
			return false;
		}
  }

  private boolean isProcessConnector(Class<?> clazz) {
  	try {
  		Object connector = clazz.newInstance();
  		return connector instanceof ProcessConnector;
  	} catch (Exception e) {
			return false;
		}
  }

  public List<ConnectorDescription> getAllConnectors(String categoryId) {
  	return getConnectors(connectors, categoryId);
  }

  public List<ConnectorDescription> getJavaConnectors(String categoryId) {
  	List<ConnectorDescription> javaConnectors = getJavaConnectors();
  	return getConnectors(javaConnectors, categoryId);
  }

  public List<ConnectorDescription> getRoleMapperConnectors(String categoryId) {
  	List<ConnectorDescription> roleMappers = getRoleMappers();
  	return getConnectors(roleMappers, categoryId);
  }

  public List<ConnectorDescription> getMapperConnectors(String categoryId) {
  	List<ConnectorDescription> mappers = getMappers();
  	return getConnectors(mappers, categoryId);
  }

  private List<ConnectorDescription> getConnectors(List<ConnectorDescription> list, String categoryId) {
    List<ConnectorDescription> connectors = new ArrayList<ConnectorDescription>();
    for (ConnectorDescription connector : list) {
      String cat = connector.getCategoryId();
      if (cat == null) {
        cat = "Other";
      }
      if (categoryId.equals(cat)) {
      	connectors.add(connector);
      }
    }
    return connectors;
  }

  public ConnectorDescription getConnector(String id) {
  	return getConnector(connectors, id);
  } 

  public ConnectorDescription getRoleMapperConnector(String id) {
  	List<ConnectorDescription> roleMappers = getRoleMappers();
  	return getConnector(roleMappers, id);
  }

  public ConnectorDescription getJavaConnector(String id) {
  	List<ConnectorDescription> javaConnectors = getJavaConnectors();
  	return getConnector(javaConnectors, id);
  }

  public ConnectorDescription getMapperConnector(String id) {
  	List<ConnectorDescription> mappers = getMappers();
  	return getConnector(mappers, id);
  }

  private ConnectorDescription getConnector(List<ConnectorDescription> list, String id) {
  	ConnectorDescription desc = null;
    int size = list.size();
    int i = 0;
    boolean found = false;
    while (i < size && !found) {
      ConnectorDescription temp = list.get(i);
      if (id.equals(temp.getId())) {
        desc = temp;
        found = true;
      }
      i++;
    }
    if (!found) {
      throw new IllegalArgumentException(id + " is not a valid identifier");
    }
    return desc;
  }

  public List<String> getAllCategories() {
    return getCategories(connectors);
  }

  public List<String> getJavaConnectorsCategories() {
  	List<ConnectorDescription> javaConnectors = getJavaConnectors();
    return getCategories(javaConnectors);
  }

  public List<String> getRoleMappersCategories() {
    List<ConnectorDescription> roleMappers = getRoleMappers();
    return getCategories(roleMappers);
  }

  public List<String> getMappersCategories() {
  	List<ConnectorDescription> mappers = getMappers();
    return getCategories(mappers);
  }
  
  public String getCategoryLabel(String categoryId) {
    return getPropertyValue(categoryId);
  }

  private List<String> getCategories(List<ConnectorDescription> list) {
    List<String> categories = new ArrayList<String>();
    boolean other = false;
    for (ConnectorDescription connector : list) {
      String categoryId = connector.getCategoryId();
      if (categoryId != null) {
        String cat = getPropertyValue(categoryId);
        if (cat != null) {
          if (!categories.contains(categoryId)) {
            categories.add(categoryId);
          }
        } else {
          other = true;
        }
      } else {
        other = true;
      }
    }
    Collections.sort(categories);
    if (other) {
      categories.add("Other");
    }
    return categories;
  }

  private String getPropertyValue(String property) {
    try {
      ResourceBundle bundle =
        ResourceBundle.getBundle(this.getClass().getName(), currentLocale);
      return bundle.getString(property);
    } catch (Exception e) {
      return null;
    }
  }

  private boolean isAnInstanceOfConnector(Class<?> c) {
    boolean isInstance = false;
    try {
      Object o = c.newInstance();
      if (o instanceof Connector) {
        isInstance = true;
      }
    } catch(Exception e) {
      return false;
    }
    return isInstance;
  }

  private boolean isAConnector(Class<?> c) {
    boolean isConnector = false;
    if (isAnInstanceOfConnector(c)) {
      int m = c.getModifiers();
      if (!Modifier.isAbstract(m)) {
        isConnector = true;
      }
    }
    return isConnector;
  }

	public void setCurrentLocale(Locale currentLocale) {
	  this.currentLocale = currentLocale;
  }

	public Locale getCurrentLocale() {
	  return currentLocale;
  }
}
