/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 * 
 * Modifed by Matthieu Chaffotte - BonitaSoft S.A.
 **/
package org.ow2.bonita.connector.core;

import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.ResourceBundle;

import org.ow2.bonita.connector.core.annotation.Category;
import org.ow2.bonita.connector.core.annotation.Checkbox;
import org.ow2.bonita.connector.core.annotation.Enumeration;
import org.ow2.bonita.connector.core.annotation.Group;
import org.ow2.bonita.connector.core.annotation.GroupInfo;
import org.ow2.bonita.connector.core.annotation.Groups;
import org.ow2.bonita.connector.core.annotation.Id;
import org.ow2.bonita.connector.core.annotation.Input;
import org.ow2.bonita.connector.core.annotation.Language;
import org.ow2.bonita.connector.core.annotation.Option;
import org.ow2.bonita.connector.core.annotation.Page;
import org.ow2.bonita.connector.core.annotation.Pages;
import org.ow2.bonita.connector.core.annotation.Radio;
import org.ow2.bonita.connector.core.annotation.Required;
import org.ow2.bonita.connector.core.annotation.Select;
import org.ow2.bonita.connector.core.annotation.Text;
import org.ow2.bonita.connector.core.annotation.Textarea;
import org.ow2.bonita.connector.core.widget.Component;
import org.ow2.bonita.connector.core.widget.Widget;

public class ConnectorDescription {

  private Class <? extends Connector> classConnector;
  private Locale currentLocale;

  public ConnectorDescription(Class<? extends Connector> classConnector) {
    this(classConnector, Locale.getDefault());
  }

  public ConnectorDescription(Class<? extends Connector> classConnector, Locale locale) {
    if (classConnector == null) {
      throw new IllegalArgumentException("The connector class cannot be null");
    } else {
      List<ConnectorError> errors = Connector.validateConnector(classConnector);
      if (!errors.isEmpty()) {
        for (ConnectorError e : errors) {
          System.out.println(e.getField() + " " + e.getError());
        }

        throw new IllegalArgumentException("The connector contains errors!" + classConnector);
      } else {
        this.classConnector = classConnector;
        currentLocale = locale;
      }
    }
  }

  /**
   * 
   * @param locale
   */
  public void setLocale(Locale locale) {
    currentLocale = locale;
  }

  /**
   * Gets the Connector Identification.
   * It checks whether the Connector has an identification
   * and returns either the Connector identification or the
   * Connector class name.
   * @return the connector identification or the Connector class name.
   */
  public String getId() {
    Id id = classConnector.getAnnotation(Id.class);
    String identification = new String();
    if (id != null) {
      identification = id.value();
    } else {
      identification = classConnector.getSimpleName();
    }
    return identification;
  }

  /**
   * Gets the Connector Description.
   * It checks whether the Connector has a description and returns either
   * the Connector Description or null
   * @return the connector description or null
   */
  public String getDescription() {
    return getResourceBundleValue("Description");
  }

  /**
   * Gets the Connector Category Id.
   * It checks whether the Connector has a category and returns either
   * the Connector Category or null
   * @return the connector category or null
   */
  public String getCategoryId() {
    Category category = classConnector.getAnnotation(Category.class);
    if (category == null) {
      return null;
    } else {
      return category.value();
    }
  }

  /**
   * Gets all the Page identifiers.
   * it checks whether the Connector has page names and returns either
   * an array of page names or an empty array.
   * @return an array of page names or an empty array
   */
  public String[] getPages() {
    Pages pages = classConnector.getAnnotation(Pages.class);
    String[] array = null;
    if (pages == null)
      array = new String[0];
    else {
      array = pages.value();
    }
    return array;
  }

  /**
   * Gets the page name
   * @param pageId the page identifier
   * @return the page description
   */
  public String getPageName(String pageId) {
    return getResourceBundleValue(pageId);
  }

  /**
   * Gets the page description
   * @param pageId the page identifier
   * @return the page description
   */
  public String getPageDescription(String pageId) {
    return getResourceBundleValue(pageId + ".description");
  }

  /**
   * Gets all the components of the connector inputs in a random order
   * @return
   */
  public List<Component> getAllInputs() {
    List<Component> componentsList = new ArrayList<Component>();
    for (Method m : classConnector.getMethods()) {
      Input input = m.getAnnotation(Input.class);
      if (input != null) {
        Component component = getWidget(m);
        componentsList.add(component);
      }
    }
    return componentsList;
  }

  /**
   * Gets all components of a connector page. If the page does not exist,
   * an Exception is raised
   * @param pageId
   * @return
   */
  public List<Component> getAllPageInputs(String pageId) {
    Pages pages = classConnector.getAnnotation(Pages.class);
    List<Component> components = new ArrayList<Component>();
    if (pages == null) {
      throw new IllegalArgumentException(classConnector.getSimpleName()
          + " does not contain any pages");
    } else {
      if (!pageExists(pages.value(), pageId)) {
        throw new IllegalArgumentException(pageId + "is not a Connector page name");
      } else {
        Map<Integer, Component> comps = new Hashtable<Integer, Component>();
        for (Method m : classConnector.getMethods()) {
          Group group = m.getAnnotation(Group.class);
          Page page = m.getAnnotation(Page.class);
          // page position
          if (group != null) {
            GroupInfo groupInfo = getPageGroupPosition(group.id());
            if (pageId.equals(groupInfo.pageId())) {
              int order = groupInfo.pageOrder();
              org.ow2.bonita.connector.core.widget.Group current = (org.ow2.bonita.connector.core.widget.Group) comps.get(order - 1);
              if (current == null) {
                String label = getResourceBundleValue(groupInfo.groupId() + ".label");
                if (label == null) {
                  label = groupInfo.groupId();
                }
                current = new org.ow2.bonita.connector.core.widget.Group(label, groupInfo.optional());
              }
              Widget component = getWidget(m);
              current.addWidget(group.order() - 1, component);
              comps.put(order - 1, current);
            }
          } else if (page != null && pageId.equals(page.page())) {
            Component component = getWidget(m);
            comps.put(page.order() - 1, component);
          }
        }
        for (int i = 0; i < comps.size(); i++) {
          components.add(comps.get(i));
        }
      }
      return components;
    }
  }

  private GroupInfo getPageGroupPosition(String groupId) {
    Groups groups = classConnector.getAnnotation(Groups.class);
    GroupInfo[] infos = groups.value();
    for (int i = 0; i < infos.length; i++) {
      if (groupId.equals(infos[i].groupId())) {
        return infos[i];
      }
    }
    throw new IllegalArgumentException(groupId + "does not exist");
  }

  private boolean pageExists(String[] pages, String pageName) {
    boolean exists = false;
    for (int i = 0; i < pages.length; i++) {
      if (pageName.equals(pages[i])) {
        exists = true;
        break;
      }
    }
    return exists;
  }

  private Widget getWidget(Method m) {
    Widget comp = null;
    String label = getInputLabel(m);
    Text text = m.getAnnotation(Text.class);
    Textarea ta = m.getAnnotation(Textarea.class);
    Radio radio = m.getAnnotation(Radio.class);
    Checkbox box = m.getAnnotation(Checkbox.class);
    Select select = m.getAnnotation(Select.class);
    Enumeration enume = m.getAnnotation(Enumeration.class);
    Required require = m.getAnnotation(Required.class);

    boolean required = false;
    if (require != null && require.value().length() == 0) {
      required = true;
    }
    String parameterName = Connector.getFieldName(m.getName());
    
    Field field = Connector.getField(classConnector, parameterName);
    Class<?> type = field.getType();

    if (text != null) {
      comp = new org.ow2.bonita.connector.core.widget.Text(
          label, parameterName, required, type, text.size(), text.maxChar());
    } else if (ta != null) {
      comp = new org.ow2.bonita.connector.core.widget.Textarea(
          label, parameterName, required, type, ta.rows(), ta.columns(), ta.maxChar(), ta.maxCharPerRow());
    } else if (radio != null) {
      comp = new org.ow2.bonita.connector.core.widget.Radio(
          label, parameterName, required, type, radio.name(), radio.value(), radio.checked());
    } else if (box != null) {
      comp = new org.ow2.bonita.connector.core.widget.Checkbox(
          label, parameterName, required, type, box.name(), box.value(), box.checked());
    } else if (select != null) {
      Option[] options = select.values();
      Map<String, String> values = new HashMap<String, String>();
      for (int i = 0; i < options.length; i++) {
        String fieldName = Connector.getFieldName(m.getName());
        String keyValue = getResourceBundleValue(fieldName + "." + options[i].id());
        if (keyValue == null) {
          keyValue = options[i].id();
        }
        values.put(keyValue, options[i].value());
      }
      org.ow2.bonita.connector.core.widget.Option option =
        new org.ow2.bonita.connector.core.widget.Option(select.top().id(), select.top().value());
      comp = new org.ow2.bonita.connector.core.widget.Select(
          label, parameterName, required, type, values, select.editable(), option);
    } else if (enume != null) {
      Option[] options = enume.values();
      Map<String, String> values = new HashMap<String, String>();
      for (int i = 0; i < options.length; i++) {
        String fieldName = Connector.getFieldName(m.getName());
        String keyValue = getResourceBundleValue(fieldName + "." + options[i].id());
        if (keyValue == null) {
          keyValue = options[i].id();
        }
        values.put(keyValue, options[i].value());
      }
      comp = new org.ow2.bonita.connector.core.widget.Enumeration(
          label, parameterName, required, type, values, enume.selectedIndices(), enume.lines(), enume.selection());
    } else {
      comp = new org.ow2.bonita.connector.core.widget.Text(
          label, parameterName, required, type, 20, 20);
    }
    return comp;
  }

  private String getInputLabel(Method m) {
    String fieldName = Connector.getFieldName(m.getName());
    String label = getResourceBundleValue(fieldName + ".label");
    if (label == null) {
      return fieldName;
    } else {
      return label;
    }
  }

  private ResourceBundle getResourceBundle() {
    Language language = classConnector.getAnnotation(Language.class);
    if (language == null) {
      return null;
    } else {
      return ResourceBundle.getBundle(language.value(), currentLocale);
    }
  }

  private String getResourceBundleValue(String key) {
    ResourceBundle messages = getResourceBundle();
    if (messages == null) {
      return null;
    }
    try {
      return messages.getString(key);
    } catch(Exception e) {
      return null;
    }
  }

  public Class<? extends Connector> getConnectorClass() {
    return classConnector;
  }

  @Override
  public boolean equals(Object object) {
    boolean equals = false;
    if (object != null && object instanceof ConnectorDescription) {
      ConnectorDescription description = (ConnectorDescription) object;
      if (this.getId().equals(description.getId())) {
        equals = true;
      }
    }
    return equals;
  }
}
