/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 * 
 * Modified by Matthieu Chaffotte - BonitaSoft S.A.
 **/
package org.ow2.bonita.connector.core;

import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.ResourceBundle;
import java.util.Map.Entry;

import org.ow2.bonita.connector.core.annotation.Category;
import org.ow2.bonita.connector.core.annotation.Checkbox;
import org.ow2.bonita.connector.core.annotation.Enumeration;
import org.ow2.bonita.connector.core.annotation.Forbidden;
import org.ow2.bonita.connector.core.annotation.Group;
import org.ow2.bonita.connector.core.annotation.GroupInfo;
import org.ow2.bonita.connector.core.annotation.Groups;
import org.ow2.bonita.connector.core.annotation.Id;
import org.ow2.bonita.connector.core.annotation.Input;
import org.ow2.bonita.connector.core.annotation.Language;
import org.ow2.bonita.connector.core.annotation.Option;
import org.ow2.bonita.connector.core.annotation.Output;
import org.ow2.bonita.connector.core.annotation.Page;
import org.ow2.bonita.connector.core.annotation.Pages;
import org.ow2.bonita.connector.core.annotation.Radio;
import org.ow2.bonita.connector.core.annotation.Required;
import org.ow2.bonita.connector.core.annotation.Select;
import org.ow2.bonita.connector.core.annotation.Text;
import org.ow2.bonita.connector.core.annotation.Textarea;
import org.ow2.bonita.expression.ExpressionEvaluator;
import org.ow2.bonita.expression.InvalidExpressionException;

public class ConnectorValidator {

  public static List<ConnectorError> validateRuntimeAnnotations(
      final Class<? extends Connector> c) {
    List<ConnectorError> errors = new ArrayList<ConnectorError>();
    ConnectorError error = null;
    for (Method m : c.getMethods()) {
      Input input = m.getAnnotation(Input.class);
      Required required = m.getAnnotation(Required.class);
      Output output = m.getAnnotation(Output.class);
      Forbidden forbidden = m.getAnnotation(Forbidden.class);
      String methodName = m.getName();
      if (input != null) {
        if (!methodName.startsWith("set")) {
          error = new ConnectorError(methodName,
              new IllegalArgumentException("Input annotation must be on a "
                  + "setter method"));
          errors.add(error);
        }
        String fieldName = Connector.getFieldName(methodName);
        if (!Connector.isFieldExist(c, fieldName)) {
          error = new ConnectorError(methodName,
              new IllegalArgumentException("Setter method name refers to an "
                  + "unknown field name"));
          errors.add(error);
        }
      }
      if (output != null) {
        if (!methodName.startsWith("get")
            && !methodName.startsWith("is")) {
          error = new ConnectorError(methodName,
              new IllegalArgumentException("Output annotation must be on a "
                  + "getter method"));
          errors.add(error);
        }
        String fieldName = Connector.getFieldName(methodName);
        if (!Connector.isFieldExist(c, fieldName)) {
          error = new ConnectorError(methodName,
              new IllegalArgumentException("Getter method name refers to an "
                  + "unknown field name"));
          errors.add(error);
        }
      }
      if (required != null) {
        if (!isWithInputOrOutput(m)) {
          error = new ConnectorError(methodName,
              new IllegalArgumentException("The required annotation must be "
                  + "added to an input or an output annotation"));
          errors.add(error);
        }
        String expression = required.value();
        if (expression.length() > 0) {
          if (expression.length() >= 0) {
            InvalidExpressionException e =
              isValidBooleanExpression(expression);
            if (e != null) {
              error = new ConnectorError(methodName, e);
              errors.add(error);
            }
          }
        }
      }
      if (forbidden != null) {
        if (!isWithInputOrOutput(m)) {
          error = new ConnectorError(methodName,
              new IllegalArgumentException("The forbidden annotation must be "
                  + "added to an input or an output annotation"));
          errors.add(error);
        }
        String expression = forbidden.value();
        if (expression.length() > 0) {
          if (expression.length() >= 0) {
            InvalidExpressionException e =
              isValidBooleanExpression(expression);
            if (e != null) {
              error = new ConnectorError(methodName, e);
              errors.add(error);
            }
          }
        }
      }
    }
    return errors;
  }

  /**
   * Checks whether the given expression is a valid boolean expression.
   * @param expression the expression to check
   * @return null if the expression is well-formed; an
   * InvalidExpressionException otherwise
   */
  protected static InvalidExpressionException isValidBooleanExpression(
      final String expression) {
    try {
      new ExpressionEvaluator(expression);
      return null;
    } catch (InvalidExpressionException e) {
      return e;
    }
  }

  public static List<ConnectorError> validateGraphicalAnnotations(
      final Class<? extends Connector> c) {
    // <Page label, list of components order>
    Map<String, List<Integer>> pagesOrder = new HashMap<String, List<Integer>>();
    List<ConnectorError> errors = new ArrayList<ConnectorError>();
    List<ConnectorError> temp = checkId(c);
    if (!temp.isEmpty()) {
      errors.addAll(temp);
    }
    temp = checkLanguage(c);
    if (!temp.isEmpty()) {
      errors.addAll(temp);
    }
    temp = checkPages(c, pagesOrder);
    if (!temp.isEmpty()) {
      errors.addAll(temp);
    }
    temp = checkCategory(c);
    if (!temp.isEmpty()) {
      errors.addAll(temp);
    }

    Map<String, List<Integer>> groupIdList = new HashMap<String, List<Integer>>();
    temp = checkGroups(c, pagesOrder, groupIdList);
    if (!temp.isEmpty()) {
      errors.addAll(temp);
    }
    Map<String, Boolean> radioList = new HashMap<String, Boolean>();
    
    ConnectorError error = null;
    for (Method m : c.getMethods()) {
      if (getWidgetNumber(m) > 1) {
        error = new ConnectorError(m.getName(),
            new IllegalArgumentException("Only one widget is possible. It can be "
                + "Text, Textarea, Radio, Checkbox, Select or Enumeration"));
        errors.add(error);
      }
      errors.addAll(checkTextWidget(m));
      errors.addAll(checkTextareaWidget(m));
      errors.addAll(checkRadioWidget(m, radioList));
      errors.addAll(checkCheckboxWidget(m));
      errors.addAll(checkSelectWidget(m));
      errors.addAll(checkEnumerationWidget(m));
      errors.addAll(checkPage(c, m, pagesOrder));
      errors.addAll(checkGroup(c, m, groupIdList));
    }
    errors.addAll(checkPagesOrder(c, pagesOrder));
    errors.addAll(checkGroupOrder(c, groupIdList));
    return errors;
  }

  private static List<ConnectorError> checkCategory(Class<? extends Connector> c) {
    Category category = c.getAnnotation(Category.class);
    List<ConnectorError> errors = new ArrayList<ConnectorError>();
    if (category != null) {
      ConnectorError error = null;
      if ("".equals(category.value().trim())) {
        error = new ConnectorError(c.getSimpleName(),
            new IllegalArgumentException("The catgory ID is empty"));
        errors.add(error);
      }
    }
    return errors;
  }

  private static List<ConnectorError> checkGroup(
      Class<? extends Connector> c, Method m, Map<String, List<Integer>> groupIdList) {
    List<ConnectorError> errors = new ArrayList<ConnectorError>();
    Group group = m.getAnnotation(Group.class);
    Groups groups = c.getAnnotation(Groups.class);
    if (group != null) {
      ConnectorError error = null;
      if (groups == null) {
        error = new ConnectorError(m.getName(),
            new IllegalArgumentException("In order to use group annotation, groups annotation is required"));
        errors.add(error);
      } else {
        if (!isWithInput(m)) {
          error = new ConnectorError(m.getName(),
              new IllegalArgumentException("The input annotation is missing because of the group annotation"));
          errors.add(error);
        }
        Page page = m.getAnnotation(Page.class);
        if (page != null) {
          error = new ConnectorError(m.getName(),
              new IllegalArgumentException("Impossible to find group and page annotation on the same method"));
          errors.add(error);
        }
      }
      if("".equals(group.id().trim())) {
        error = new ConnectorError(m.getName(),
            new IllegalArgumentException("A group Id is empty"));
        errors.add(error);
      } else if (!groupIdList.containsKey(group.id())){
        error = new ConnectorError(m.getName(),
            new IllegalArgumentException(group.id() + " does not exist in Groups annotation"));
        errors.add(error);
      } else {
        if (group.order() < 1) {
          error = new ConnectorError(m.getName(),
              new IllegalArgumentException("A group order must start from 1"));
          errors.add(error);
        } else {
          List<Integer> list = groupIdList.get(group.id());
          if (list.contains(group.order())) {
            error = new ConnectorError(m.getName(),
                new IllegalArgumentException(group.order() + " is already taken by another method"));
            errors.add(error);
          } else {
            list.add(group.order());
            groupIdList.put(group.id(), list);
          }
        }
      }
    }
    return errors;
  }

  private static List<ConnectorError> checkLanguage(Class<? extends Connector> c) {
  	List<ConnectorError> errors = new ArrayList<ConnectorError>();
  	Language language = c.getAnnotation(Language.class);
  	if (language != null) {
  		ConnectorError error = null;
  		String baseName = language.value();
  		if (baseName.length() == 0) {
  			error = new ConnectorError(c.getSimpleName(),
            new IllegalArgumentException("The language value cannot be empty"));
        errors.add(error);
  		} else {
  			ResourceBundle resource = getDefaultLanguage(baseName);
  			if (resource == null) {
  				error = new ConnectorError(c.getSimpleName(),
              new IllegalArgumentException("The default properties file does not exist in " + baseName));
  				errors.add(error);
  			}
  		}
  	}
	  return errors;
  }

  private static ResourceBundle getDefaultLanguage(String baseName) {
	  try {
	  	return ResourceBundle.getBundle(baseName, Locale.getDefault());
	  } catch (Exception e) {
			return null;
		}
  }

  private static List<ConnectorError> checkGroups(Class<? extends Connector> c,
      Map<String, List<Integer>> pagesOrder, Map<String, List<Integer>> groupIdList) {
    List<ConnectorError> errors = new ArrayList<ConnectorError>();
    Groups groups = c.getAnnotation(Groups.class);
    if (groups != null) {
      GroupInfo[] groupsInfo = groups.value();
      for (int i = 0; i < groupsInfo.length; i++) {
        ConnectorError error = null;
        GroupInfo current = groupsInfo[i];
        //checking groupID
        if ("".equals(current.groupId().trim())) {
          error = new ConnectorError("Groups",
              new IllegalArgumentException("A groupInfo Id is empty"));
          errors.add(error);
        } else {
          if (groupIdList.containsKey(current.groupId())) {
            error = new ConnectorError(current.groupId(),
                new IllegalArgumentException("this Id is already taken by another group"));
            errors.add(error);
          } else {
            groupIdList.put(current.groupId(), new ArrayList<Integer>());
          }
        }
        //checking pageID
        if("".equals(current.pageId().trim())) {
          error = new ConnectorError("Groups",
              new IllegalArgumentException("A page Id is empty"));
          errors.add(error);
        } else if (!pagesOrder.containsKey(current.pageId())) {
          error = new ConnectorError("Groups",
              new IllegalArgumentException(current.pageId() + " does not refer to an existing page"));
          errors.add(error);
        }
        //checking pageOrder
        if (current.pageOrder() < 1) {
          error = new ConnectorError("Groups",
              new IllegalArgumentException("A page order must start from 1"));
          errors.add(error);
        }
        //OK
        if (error == null) {
          List<Integer> list = pagesOrder.get(groupsInfo[i].pageId());
          Integer temp = new Integer(groupsInfo[i].pageOrder());
          if (list.contains(temp)) {
            error = new ConnectorError("Groups",
                new IllegalArgumentException("The position " + temp + " has already taken by another parameter"));
            errors.add(error);
          } else {
            list.add(temp);
            pagesOrder.put(groupsInfo[i].pageId(), list);
          }
        }
      }
    }
    return errors;
  }

  private static List<ConnectorError> checkGroupOrder(
      Class<? extends Connector> c, Map<String, List<Integer>> groupIdList) {
    List<ConnectorError> errors = new ArrayList<ConnectorError>();
    ConnectorError error = null;
    for (Entry<String, List<Integer>> entry : groupIdList.entrySet()) {
      List<Integer> values = entry.getValue();
      if (values.isEmpty()) {
        error = new ConnectorError(entry.getKey(),
            new IllegalArgumentException("does not contain any group reference"));
        errors.add(error);
      } else {
        for (int j = 1; j <= values.size(); j++) {
          if (!values.contains(j)) {
            error = new ConnectorError(entry.getKey(),
                new IllegalArgumentException(j + " is missing in the page order"));
            errors.add(error);
          }
        }
      }
    }
    return errors;    
  }

	private static List<ConnectorError> checkPagesOrder(Class<? extends Connector> c,
      Map<String, List<Integer>> pagesOrder) {
    List<ConnectorError> errors = new ArrayList<ConnectorError>();
    Pages pages = c.getAnnotation(Pages.class);
    if (pages != null) {
      int pageNumber = pagesOrder.size();
      String[] pagesIds = pages.value();
      ConnectorError error = null;
      for (int i = 0; i < pageNumber; i++) {
        String pageId = pagesIds[i];
        if (pageId.length() != 0) {
          List<Integer> pageOrder = pagesOrder.get(pagesIds[i]);
          if (pageOrder.isEmpty()) {
            error = new ConnectorError(pageId,
                new IllegalArgumentException("does not contain any page reference"));
            errors.add(error);
          } else {
            for (int j = 1; j <= pageOrder.size(); j++) {
              if (!pageOrder.contains(j)) {
                error = new ConnectorError(pageId,
                    new IllegalArgumentException(j + " is missing in the page order"));
                errors.add(error);
              }
            }
          }
        }
      } 
    }
    return errors;
  }

  private static List<ConnectorError> checkPages(Class<? extends Connector> c,
      Map<String, List<Integer>> pagesOrder) {
    List<ConnectorError> errors = new ArrayList<ConnectorError>();
    ConnectorError error = null;
    Pages pages = c.getAnnotation(Pages.class);
    
    if (pages != null) {
      String[] pageNames = pages.value();
      List<String> names = new ArrayList<String>();
      if (pageNames.length == 0) {
        error = new ConnectorError(c.getSimpleName(),
            new IllegalArgumentException("The Pages value cannot be empty"));
        errors.add(error);
      } else {
        for (int i = 0; i < pageNames.length; i++) {
          if (pageNames[i].length() == 0) {
            error = new ConnectorError(c.getSimpleName(),
                new IllegalArgumentException("A value of Pages cannot be empty"));
            errors.add(error);
          } else {
            if (names.contains(pageNames[i])) {
              error = new ConnectorError(c.getSimpleName(),
                  new IllegalArgumentException(pageNames[i] + " already exists in Pages"));
              errors.add(error);
            }
            else {
              names.add(pageNames[i]);
              pagesOrder.put(pageNames[i], new ArrayList<Integer>());
            }
          }
        }
      }
    }
    return errors;
  }

  private static List<ConnectorError> checkId(Class<? extends Connector> c) {
    List<ConnectorError> errors = new ArrayList<ConnectorError>();
    Id id = c.getAnnotation(Id.class);
    if (id != null && id.value().length() == 0) {
      ConnectorError error = new ConnectorError(c.getSimpleName(),
          new IllegalArgumentException("The Id value cannot be emtpy"));
      errors.add(error);
    }
    return errors;
  }

  private static List<ConnectorError> checkPage(Class<? extends Connector> c, 
      Method m, Map<String, List<Integer>> pagesOrder) {
    List<ConnectorError> errors = new ArrayList<ConnectorError>();
    ConnectorError error = null;
    Page page = m.getAnnotation(Page.class);
    Pages pages = c.getAnnotation(Pages.class);
    if (page != null) {
      if (pages == null) {
        error = new ConnectorError(m.getName(),
            new IllegalArgumentException("In order to use page annotation, pages annotation is required"));
        errors.add(error);
      } else {
        if (!isWithInput(m)) {
          error = new ConnectorError(m.getName(),
              new IllegalArgumentException("The input annotation is missing because of the page annotation"));
          errors.add(error);
        }
        String pageId = page.page();
        String[] pageIds = pages.value();
        if (pageId.length() == 0) {
          error = new ConnectorError(m.getName(),
              new IllegalArgumentException("A page id cannot be empty"));
          errors.add(error);
        } else if (!pageIdExists(pageId, pageIds)){
          error = new ConnectorError(m.getName(),
              new IllegalArgumentException(pageId + " does not exist in pages annotation"));
          errors.add(error);
        } else if (page.order() <= 0) {
          error = new ConnectorError(m.getName(),
              new IllegalArgumentException("A page order must start from 1"));
          errors.add(error);
        } else {
          List<Integer> order = pagesOrder.get(page.page());
          Integer temp = new Integer(page.order());
          if (!order.contains(temp)) {
            order.add(temp);
            pagesOrder.put(page.page(), order);
          } else {
            error = new ConnectorError(m.getName(),
                new IllegalArgumentException("The position " + page.order() + " has already taken by another parameter"));
            errors.add(error);
          } 
        } 
      }
    } else { //page is null
      Input input = m.getAnnotation(Input.class);
      Group group = m.getAnnotation(Group.class);
      if (pages != null && input != null && group == null) {
        error = new ConnectorError(m.getName(),
            new IllegalArgumentException("page or group annotation must be defined because of the pages annotation"));
        errors.add(error);
      }
    }
    return errors;
  }

  private static boolean pageIdExists(String pageId, String[] pageIds) {
    boolean exists = false;
    for (String id : pageIds) {
      if (id.equals(pageId)) {
        exists = true;
        break;
      }
    }
    return exists;
  }

  private static List<ConnectorError> checkEnumerationWidget(Method m) {
    List<ConnectorError> errors = new ArrayList<ConnectorError>();
    ConnectorError error = null;
    Enumeration enume = m.getAnnotation(Enumeration.class);
    if (enume != null) {
      if (!isWithInput(m)) {
        error = new ConnectorError(m.getName(),
            new IllegalArgumentException("The input annotation is missing because of the enumeration annotation"));
        errors.add(error);
      }
      if (enume.lines() < 1) {
        error = new ConnectorError(m.getName(),
            new IllegalArgumentException("The line number cannot be less than 1"));
        errors.add(error);
      }
      Option[] options = enume.values();
      if (options.length == 0) {
        error = new ConnectorError(m.getName(),
            new IllegalArgumentException("The enumeration values cannot be empty"));
        errors.add(error);
      } else {
        List<String> labels = new ArrayList<String>();
        for (int i = 0; i < options.length; i++) {
          String label = options[i].id();
          if (label.length() == 0) {
            error = new ConnectorError(m.getName(),
                new IllegalArgumentException("An option label cannot be empty"));
            errors.add(error);
          }
          if (options[i].value().length() == 0) {
            error = new ConnectorError(m.getName(),
                new IllegalArgumentException("An option value cannot be empty"));
            errors.add(error);
          }
          if (labels.contains(label)) {
            error = new ConnectorError(m.getName(),
                new IllegalArgumentException("An option label is unique in a select"));
            errors.add(error);
          } else {
            labels.add(label);
          }
        }
      }
      int[] indices = enume.selectedIndices();
      List<Integer> integers = new ArrayList<Integer>();
      if (indices.length > options.length) {
        error = new ConnectorError(m.getName(),
            new IllegalArgumentException("Impossible to have more selected indices than available options"));
        errors.add(error);
      }
      for (int i = 0; i < indices.length; i++) {
        if (indices[i] < 0) {
          error = new ConnectorError(m.getName(),
              new IllegalArgumentException("Indices cannot contain a nevative value"));
          errors.add(error);
        }
        if (indices[i] > options.length) {
          error = new ConnectorError(m.getName(),
              new IllegalArgumentException("An indice cannot be greater than the enumeration"));
          errors.add(error);
        }
        if (integers.contains(indices[i])) {
          error = new ConnectorError(m.getName(),
              new IllegalArgumentException("It is not allowed to have two identical indices"));
          errors.add(error);
        } else if(integers.contains(0) && indices.length > 1) {
          error = new ConnectorError(m.getName(),
              new IllegalArgumentException("Either 0 or selected indices upper than 0 are allowed, not both"));
          errors.add(error);
        }else{
          integers.add(indices[i]);
        }
      }
    }
    return errors;
  }

  private static List<ConnectorError> checkSelectWidget(Method m) {
    List<ConnectorError> errors = new ArrayList<ConnectorError>();
    ConnectorError error = null;
    Select select = m.getAnnotation(Select.class);
    if (select != null) {
      if (!isWithInput(m)) {
        error = new ConnectorError(m.getName(),
            new IllegalArgumentException("The input annotation is missing because of the select annotation"));
        errors.add(error);
      }
      Option[] options = select.values();
      if (options.length == 0) {
        error = new ConnectorError(m.getName(),
            new IllegalArgumentException("The select values cannot be empty"));
        errors.add(error);
      } else {
        List<String> labels = new ArrayList<String>();
        for (int i = 0; i < options.length; i++) {
          String label = options[i].id();
          if (label.length() == 0) {
            error = new ConnectorError(m.getName(),
                new IllegalArgumentException("An option label cannot be empty"));
            errors.add(error);
          }
          if (options[i].value().length() == 0) {
            error = new ConnectorError(m.getName(),
                new IllegalArgumentException("An option value cannot be empty"));
            errors.add(error);
          }
          if (labels.contains(label)) {
            error = new ConnectorError(m.getName(),
                new IllegalArgumentException("An option label is unique in a select"));
            errors.add(error);
          } else {
            labels.add(label);
          }
        }
      }
      //top can have an empty label and empty value
    }
    return errors;
  }

  private static List<ConnectorError> checkCheckboxWidget(Method m) {
    List<ConnectorError> errors = new ArrayList<ConnectorError>();
    ConnectorError error = null;
    Checkbox check = m.getAnnotation(Checkbox.class);
    if (check != null) {
      if (!isWithInput(m)) {
        error = new ConnectorError(m.getName(),
            new IllegalArgumentException("The input annotation is missing because of the checkbox annotation"));
        errors.add(error);
      }
      if (check.name().length() == 0) {
        error = new ConnectorError(m.getName(),
            new IllegalArgumentException("The checkbox name cannot be emtpy"));
        errors.add(error);
      }
      if (check.value().length() == 0) {
        error = new ConnectorError(m.getName(),
            new IllegalArgumentException("The checkbox value cannot be emtpy"));
        errors.add(error);
      }
    }
    return errors;
  }

  private static List<ConnectorError> checkRadioWidget(Method m, Map<String, Boolean> radioList) {
    List<ConnectorError> errors = new ArrayList<ConnectorError>();
    ConnectorError error = null;
    Radio radio = m.getAnnotation(Radio.class);
    if (radio != null) {
      if (!isWithInput(m)) {
        error = new ConnectorError(m.getName(),
            new IllegalArgumentException("The input annotation is missing because of the radio annotation"));
        errors.add(error);
      }
      if (radio.name().length() == 0) {
        error = new ConnectorError(m.getName(),
            new IllegalArgumentException("The name of the radio button cannot be emtpy"));
        errors.add(error);
      }
      if (radio.value().length() == 0) {
        error = new ConnectorError(m.getName(),
            new IllegalArgumentException("The value of the radio button cannot be emtpy"));
        errors.add(error);
      }
      if (radioList.containsKey(radio.name())) {
        if (radioList.get(radio.name())) {
          if (radio.checked()) {
            String err = "Only one radio button can be checked in the group "
              + radio.name();
            error = new ConnectorError(m.getName(), new IllegalArgumentException(err));
            errors.add(error);
          }
        } else {
          radioList.put(radio.name(), radio.checked());
        }
      } else {
        radioList.put(radio.name(), radio.checked());
      }
    }
    return errors;
  }

  private static List<ConnectorError> checkTextareaWidget(Method m) {
    List<ConnectorError> errors = new ArrayList<ConnectorError>();
    ConnectorError error = null;
    Textarea text = m.getAnnotation(Textarea.class);
    if (text != null) {
      if (!isWithInput(m)) {
        error = new ConnectorError(m.getName(),
            new IllegalArgumentException("The input annotation is missing because of the textarea annotation"));
        errors.add(error);
      }
      if (text.columns() < 1) {
        error = new ConnectorError(m.getName(),
            new IllegalArgumentException("The column number of the textarea cannot be less than 1"));
        errors.add(error);
      }
      if (text.rows() < 1) {
        error = new ConnectorError(m.getName(),
            new IllegalArgumentException("The row number of the textarea cannot be less than 1"));
        errors.add(error);
      }
      if (text.maxChar() < 1) {
        error = new ConnectorError(m.getName(),
            new IllegalArgumentException("The maximum number of characters cannot be less than 1"));
        errors.add(error);
      }
      if (text.maxCharPerRow() < 1) {
        error = new ConnectorError(m.getName(),
            new IllegalArgumentException("The maximum number of characters per row cannot be less than 1"));
        errors.add(error);
      }
    }
    return errors;
  }

  private static List<ConnectorError> checkTextWidget(Method m) {
    List<ConnectorError> errors = new ArrayList<ConnectorError>();
    ConnectorError error = null;
    Text text = m.getAnnotation(Text.class);
    if (text != null) {
      if (!isWithInput(m)) {
        error = new ConnectorError(m.getName(),
            new IllegalArgumentException("The input annotation is missing because of the text annotation"));
        errors.add(error);
      }
      if (text.size() < 1) {
        error = new ConnectorError(m.getName(),
            new IllegalArgumentException("The text size cannot be less than 1"));
        errors.add(error);
      }
      if (text.maxChar() < 1) {
        error = new ConnectorError(m.getName(),
            new IllegalArgumentException("The maximum number of characters cannot be less than 1"));
        errors.add(error);
      }
    }
    return errors;
  }

  private static boolean isWithInput(Method m) {
    boolean with = false;
    Input input = m.getAnnotation(Input.class);
    if (input != null) {
      with = true;
    }
    return with;
  }

  private static boolean isWithInputOrOutput(Method m) {
    boolean with = false;
    Input input = m.getAnnotation(Input.class);
    Output output = m.getAnnotation(Output.class);
    if (input != null || output != null) {
      with = true;
    }
    return with;
  }

  private static int getWidgetNumber(Method m) {
    int count = 0;
    if (m.getAnnotation(Text.class) != null) {
      count++;
    }
    if (m.getAnnotation(Textarea.class) != null) {
      count++;
    }
    if (m.getAnnotation(Radio.class) != null) {
      count++;
    }
    if (m.getAnnotation(Checkbox.class) != null) {
      count++;
    }
    if (m.getAnnotation(Select.class) != null) {
      count++;
    }
    if (m.getAnnotation(Enumeration.class) != null) {
      count++;
    }
    return count;
  }
}
