/**
 * Copyright (C) 2009  BonitaSoft S.A..
 * BonitaSoft, 31 rue Gustave Eiffel - 38000 Grenoble
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.bonita.connector.impl.java;

import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.List;

import org.ow2.bonita.connector.core.Connector;
import org.ow2.bonita.connector.core.ConnectorError;
import org.ow2.bonita.connector.core.annotation.Id;
import org.ow2.bonita.connector.core.annotation.Input;
import org.ow2.bonita.connector.core.annotation.Required;
import org.ow2.bonita.connector.core.annotation.Text;

/**
 * 
 * @author Matthieu Chaffotte
 *
 */
@Id("Java Connector")
public class JavaConnector extends Connector {

  private String className;

  private String mainMethodName;
  private Object[] mainMethodParameters;
  private Class<?>[] mainMethodParametersClasses;

  private List<String> methodName;
  private List<Object[]> methodParameters;
  private List<Class<?>[]> methodsParametersClasses;

  public JavaConnector() {
    methodName = new ArrayList<String>();
    methodParameters = new ArrayList<Object[]>();
    methodsParametersClasses = new ArrayList<Class<?>[]>();
  }

  @Input @Required
  @Text(size=30,maxChar=100)
  public void setClassName(String className) {
    this.className = className;
  }

  @Input @Required
  @Text(size=30,maxChar=50)
  public void setMainMethodName(String methodName) {
    this.mainMethodName = methodName;
  }

  @Input
  public void setMainMethodParameters(ArrayList<Object> parameters) {
    if (parameters == null) {
      mainMethodParameters = null;
      mainMethodParametersClasses = null;
    } else {
      mainMethodParameters = new Object[parameters.size()];
      for (int i = 0; i < parameters.size(); i++) {
        mainMethodParameters[i] = parameters.get(i);
      }
      
      mainMethodParametersClasses = new Class[parameters.size()];
      for (int i = 0; i < parameters.size(); i++) {
        Object parameter = parameters.get(i);
        if (parameter == null) {
          mainMethodParametersClasses[i] = null;
        } else {
          mainMethodParametersClasses[i] = parameters.get(i).getClass();
        }
      }
    }
  }

  @Input
  public void setMethodName(String name, ArrayList<Object> parameters) {
    methodName.add(name);
    if (parameters == null) {
      methodParameters.add(new Object[0]);
      methodsParametersClasses.add(new Class[0]);
    } else {
      Object[] params = parameters.toArray();
      methodParameters.add(params);
      Class<?>[] temp = new Class[params.length];
      for (int i = 0; i < params.length; i++) {
        Object parameter = params[i];
        if (parameter == null) {
          temp[i] = null;
        } else {
          temp[i] = params[i].getClass();
        }
      }
      methodsParametersClasses.add(temp);
    }
  }

  @Override
  protected void executeConnector() throws Exception {
    Class<?> connectorClass = Class.forName(className);
    Object connector = connectorClass.newInstance();
    for (int i = 0; i < methodName.size(); i++) {
      Method setter = connectorClass.getMethod(methodName.get(i), methodsParametersClasses.get(i));
      setter.invoke(connector, methodParameters.get(i));
    }
    Method start = connectorClass.getMethod(mainMethodName, mainMethodParametersClasses);
    start.invoke(connector, mainMethodParameters);
  }

  @Override
  protected List<ConnectorError> validateValues() {
    List<ConnectorError> errors = new ArrayList<ConnectorError>();
    if (className.length() == 0) {
      errors.add(new ConnectorError("className",
          new IllegalArgumentException("is empty")));
    } else if (!classExists()) {
      errors.add(new ConnectorError("className",
          new IllegalArgumentException("is not a real class")));
    } else {
      if (mainMethodName.length() == 0) {
        errors.add(new ConnectorError("methodName",
            new IllegalArgumentException("is empty")));
      } else if (!methodExists(mainMethodName, mainMethodParametersClasses)) {
        errors.add(new ConnectorError("methodName",
            new IllegalArgumentException("is not in the class using parameters: " + mainMethodParametersClasses)));
      }
      // Parameters
      if (methodName.size() > 0) {
        for (int i = 0; i < methodName.size(); i++) {
          String methodN = methodName.get(i);
          if (methodN == null) {
            errors.add(new ConnectorError("", new IllegalArgumentException("A method name is null")));
          } else if (methodN.length() == 0) {
            errors.add(new ConnectorError(methodN, new IllegalArgumentException("is empty")));
          } else if (!methodExists(methodN, methodsParametersClasses.get(i))) {
            errors.add(new ConnectorError(methodN,
                new IllegalArgumentException("is not in the class using parameters: " + methodsParametersClasses.get(i))));
          }
        }
      }
    }
    return errors;
  }

  private boolean methodExists(String method, Class<?>[] parameters) {
    try {
      Class<?> connectorClass = Class.forName(className);
      connectorClass.getMethod(method, parameters);
      return true;
    } catch (Exception e) {
      return false;
    }
  }

  private boolean classExists() {
    try {
      Class.forName(className);
      return true;
    } catch (Exception e) {
      return false;
    }
  }
}
