/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.bonita.connector.impl.ldap;

import java.io.File;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.List;
import java.util.StringTokenizer;

import javax.naming.Context;
import javax.naming.NamingEnumeration;
import javax.naming.directory.Attribute;
import javax.naming.directory.Attributes;
import javax.naming.directory.SearchControls;
import javax.naming.directory.SearchResult;
import javax.naming.ldap.InitialLdapContext;
import javax.naming.ldap.LdapContext;
import javax.naming.ldap.StartTlsRequest;
import javax.naming.ldap.StartTlsResponse;

import org.ow2.bonita.connector.core.Connector;
import org.ow2.bonita.connector.core.ConnectorError;
import org.ow2.bonita.connector.core.annotation.Category;
import org.ow2.bonita.connector.core.annotation.Id;
import org.ow2.bonita.connector.core.annotation.Input;
import org.ow2.bonita.connector.core.annotation.Language;
import org.ow2.bonita.connector.core.annotation.Option;
import org.ow2.bonita.connector.core.annotation.Output;
import org.ow2.bonita.connector.core.annotation.Page;
import org.ow2.bonita.connector.core.annotation.Pages;
import org.ow2.bonita.connector.core.annotation.Required;
import org.ow2.bonita.connector.core.annotation.Select;
import org.ow2.bonita.connector.core.annotation.Text;

/**
 * This connector provides an LDAP service of querying directory service.
 * This connector does the search operation.
 * @author chaffotm
 *
 */
@Id("Ldap")
@Category("Database")
@Pages({"server", "search"})
@Language("org.ow2.bonita.connector.impl.ldap.LdapConnector")
public class LdapConnector extends Connector {

  /**
   * The host name of the directory service.
   */
  private String host = null;

  /**
   * The port number of the directory service.
   */
  private int port = 389;

  /**
   * The protocol used by the directory service. It can be
   * LDAP, LDAPS(SSL) or TLS.
   */
  private LdapProtocol protocol = null;

  /**
   * The user name if authentication is needed.
   */
  private String userName = null;

  /**
   * The password if authentication is needed.
   */
  private String password = null;

  private String certificatePath;

  /**
   * 
   */
  private String baseObject = null;
  
  /**
   * 
   */
  private LdapScope scope = LdapScope.BASE;
  private String filter = null;
  private LdapDereferencingAlias derefAliases =
      LdapDereferencingAlias.ALWAYS;
  private String[] attributes = null;
  private long sizeLimit = 0;
  private int timeLimit = 0;
  private String referralHandling = "ignore";

  //output
  private List<List<LdapAttribute>> result =
    new ArrayList<List<LdapAttribute>>();

  private String getHost() {
    return host;
  }

  private Integer getPort() {
    return port;
  }

  public LdapProtocol getProtocol() {
    return protocol;
  }

  public String getUserName() {
    return userName;
  }

  public String getPassword() {
    return password;
  }

  public String getCertificatePath() {
    return certificatePath;
  }

  public String getBaseObject() {
    return baseObject;
  }

  public LdapScope getScope() {
    return scope;
  }

  public String getFilter() {
    return filter;
  }

  public LdapDereferencingAlias getDerefAliases() {
    return derefAliases;
  }

  public String[] getAttributes() {
    return attributes;
  }

  public long getSizeLimit() {
    return sizeLimit;
  }

  public int getTimeLimit() {
    return timeLimit;
  }

  public List<List<LdapAttribute>> getListResult() {
    return result;
  }
  
  @Output
  public String getResult() {
    StringBuilder res = new StringBuilder();
    int listSize = result.size();
    int i = 0;
    for (List<LdapAttribute> elems : result) {
      int mapSize = elems.size();
      int j = 0;
      for (LdapAttribute elem : elems) {
        res.append(elem.getName());
        res.append(Connector.KEY_VALUE_SEPARATOR);
        res.append(elem.getValue());
        j++;
        if (j < mapSize) {
          res.append(Connector.ATTRIBUTE_SEPARATOR);
        }
      }
      i++;
      if (i < listSize) {
        res.append(Connector.OBJECT_SEPARATOR);
      }
    }
    return res.toString();
  }

  public String getReferralHandling() {
    return referralHandling;
  }
 
  @Input @Required
  @Page(page="server",order=1)
  public void setHost(String host) {
    this.host = host;
  }

  @Input @Required
  @Text(size=5,maxChar=5) @Page(page="server",order=2)
  public void setPort(Long port) {
    if (port == null) {
      setPort(Integer.MIN_VALUE);
    } else {
      setPort(port.intValue());
    }
  }
  
  public void setPort(int port) {
    this.port = port;
  }
  
  public void setProtocol(LdapProtocol protocol) {
    this.protocol = protocol;
  }

  @Input @Required
  @Page(page="server",order=3)
  @Select(editable=false,top=@Option(id="ldap",value="LDAP"),
  		values={@Option(id="ldap",value="LDAP"),@Option(id="ldaps",value="LDAPS"),@Option(id="tls",value="TLS")})
  public void setProtocol(String protocol) {
    this.protocol = LdapProtocol.LDAPS;
    if (protocol != null) {
      protocol = protocol.toUpperCase();
      if (protocol.equals(LdapProtocol.TLS.toString())) {
        this.protocol = LdapProtocol.TLS;
      } else if (protocol.equals(LdapProtocol.LDAP.toString())) {
        this.protocol = LdapProtocol.LDAP;
      }
    }
  }

  @Input @Required("password")
  @Page(page="server",order=4)
  public void setUserName(String userName) {
    this.userName = userName;
  }

  @Input @Required("userName")
  @Page(page="server",order=5)
  public void setPassword(String password) {
    this.password = password;
  }

  /*@Input
  public void setCertificatePath(String certificatePath) {
    this.certificatePath = certificatePath;
  }*/

  @Input @Required
  @Page(page="search",order=1) @Text(size=25,maxChar=50)
  public void setBaseObject(String baseObject) {
    this.baseObject = baseObject;
  }

  @Input @Required
  @Page(page="search",order=2)
  @Select(editable=false,top=@Option(id="sub",value="SUBTREE"),
  		values={@Option(id="sub",value="SUBTREE"),@Option(id="base",value="BASE"),@Option(id="one",value="ONELEVEL")})
  public void setScope(LdapScope scope) {
    this.scope = scope;
  }

  public void setScope(String scope) {
    this.scope = LdapScope.ONELEVEL;
    if (scope != null) {
      scope = scope.toUpperCase();
      if (scope.equals(LdapScope.BASE.toString())) {
        this.scope = LdapScope.BASE;
      } else if (scope.equals(LdapScope.SUBTREE.toString())) {
        this.scope = LdapScope.SUBTREE;
      }
    }
  }

  @Input @Required @Text(size=20,maxChar=100)
  @Page(page="search",order=3)
  public void setFilter(String filter) {
    this.filter = filter;
  }

  @Input @Required
  @Page(page="search",order=8)
  @Select(editable=false,top=@Option(id="alw",value="ALWAYS"),
  		values={@Option(id="alw",value="ALWAYS"),@Option(id="nvr",value="NEVER")
  	,@Option(id="find",value="FINDING"),@Option(id="search",value="SEARCHING")})
  public void setDerefAliases(LdapDereferencingAlias derefAliases) {
    this.derefAliases = derefAliases;
  }

  public void setDerefAliases(String derefAliases) {
    this.derefAliases = LdapDereferencingAlias.ALWAYS;
    if (derefAliases.equals("never")) {
      this.derefAliases = LdapDereferencingAlias.NEVER;
    } else if (derefAliases.equals("finding")) {
      this.derefAliases = LdapDereferencingAlias.FINDING;
    } else if (derefAliases.equals("searching")) {
      this.derefAliases = LdapDereferencingAlias.SEARCHING;
    }
  }

  @Input
  @Page(page="search",order=4)
  public void setAttributes(String[] attributes) {
    this.attributes = attributes;
  }

  public void setAttributes(String attributes) {
    StringTokenizer tokens =
      new StringTokenizer(attributes, Connector.ATTRIBUTE_SEPARATOR);
    int i = 0;
    this.attributes = new String[tokens.countTokens()];
    while(tokens.hasMoreElements()) {
      this.attributes[i] = (String) tokens.nextElement();
      i++;
    }
  }

  @Input
  @Page(page="search",order=5)
  public void setSizeLimit(long sizeLimit) {
    this.sizeLimit = sizeLimit;
  }

  public void setSizeLimit(Long sizeLimit) {
    if (sizeLimit == null) {
      setSizeLimit(Long.MIN_VALUE);
    } else {
      setSizeLimit(sizeLimit.longValue());
    }
  }

  /**
   * Sets the time-limit during a search in seconds.
   * @param timeLimit
   */
  @Input
  @Page(page="search",order=6)
  public void setTimeLimit(int timeLimit) {
    this.timeLimit = timeLimit;
  }

  /**
   * Sets the time-limit during a search in seconds.
   * @param timeLimit
   */
  public void setTimeLimit(Long timeLimit) {
    if (timeLimit == null) {
      setTimeLimit(Integer.MIN_VALUE);
    } else {
      setTimeLimit(timeLimit.intValue());
    }
  }

  @Input @Required
  @Page(page="search",order=7)
  public void setReferralHandling(String referralHandling) {
    this.referralHandling = referralHandling;
  }


  private Hashtable<String, String> getEnvironment() {
    Hashtable<String, String> environment = new Hashtable<String, String>();
    environment.put(Context.INITIAL_CONTEXT_FACTORY,
        "com.sun.jndi.ldap.LdapCtxFactory");
    environment.put(Context.PROVIDER_URL,
        "ldap://" + getHost() + ":" + getPort());
    if (getProtocol().equals(LdapProtocol.LDAPS)) {
      environment.put(Context.SECURITY_PROTOCOL, "ssl");
      
      /*if (getCertificatePath() != null) {
        environment.put("java.naming.ldap.factory.socket",
        "talend.ssl.AdvancedSocketFactory");
        talend.ssl.AdvancedSocketFactory.setCertStorePath(getCertificatePath());
      }*/
      
      
    }
    if (!getProtocol().equals(LdapProtocol.TLS)
        && getUserName() != null && getPassword() != null) {
      environment.put(Context.SECURITY_AUTHENTICATION, "simple");
      environment.put(Context.SECURITY_PRINCIPAL, getUserName());
      environment.put(Context.SECURITY_CREDENTIALS, getPassword());
    } else {
      environment.put(Context.SECURITY_AUTHENTICATION, "none");
    }
    environment.put("java.naming.ldap.derefAliases", getDerefAliases().toString().toLowerCase());
    environment.put(Context.REFERRAL, getReferralHandling());
    return environment;
  }
  
  @Override
  protected void executeConnector() throws Exception {
    Hashtable<String, String> env = getEnvironment();
    LdapContext ctx = new InitialLdapContext(env,null);

    StartTlsResponse response = null;
    if (getProtocol() == LdapProtocol.TLS) {
      StartTlsRequest request = new StartTlsRequest();
      response = (StartTlsResponse) ctx.extendedOperation(request);
      response.negotiate();
      if (getUserName() != null && getPassword() != null) {
        ctx.addToEnvironment(Context.SECURITY_AUTHENTICATION, "simple");
        ctx.addToEnvironment(Context.SECURITY_PRINCIPAL, getUserName());
        ctx.addToEnvironment(Context.SECURITY_CREDENTIALS, getPassword());
      }
    }
    SearchControls ctls = new SearchControls();
    ctls.setTimeLimit(getTimeLimit() * 1000);
    ctls.setCountLimit(getSizeLimit());
    ctls.setReturningAttributes(getAttributes());
    ctls.setSearchScope(getScope().value());
    NamingEnumeration<SearchResult> answer =
      ctx.search(getBaseObject(), getFilter(), ctls);
    long count = getSizeLimit();
    // count is useful in case of the size-limit is defined
    // the search method does not care about size-limit. It returns all entries
    // which match with the filter.
    if (count == 0) {
      count = Long.MAX_VALUE;
    }
    result = new ArrayList<List<LdapAttribute>>();
    while (count > 0 && answer.hasMore()) {
      SearchResult sr = answer.next();
      count--;
      Attributes attribs = sr.getAttributes();
      NamingEnumeration<? extends Attribute> enume = attribs.getAll();
      List<LdapAttribute> elements = new ArrayList<LdapAttribute>();
      while (enume.hasMore()) {
        Attribute attribute = enume.next();
        NamingEnumeration<?> all=  attribute.getAll();
        while (all.hasMore()) {
          Object key = all.next();
          elements.add(new LdapAttribute(attribute.getID(), key.toString()));
        }
      }
      if (!elements.isEmpty()) {
        result.add(elements);
      }
    }
    if (getProtocol() == LdapProtocol.TLS) {
      response.close();
    }
    ctx.close();
  }

  @Override
  protected List<ConnectorError> validateValues() {
    List<ConnectorError> errors = new ArrayList<ConnectorError>();
    if (getHost().length() == 0) {
      errors.add(new ConnectorError("host",
          new IllegalArgumentException("cannot be empty!")));
    }
    if (getPort() < 0) {
      errors.add(new ConnectorError("port",
          new IllegalArgumentException("cannot be less than 0!")));
    } else if (getPort() > 65535) {
      errors.add(new ConnectorError("port",
          new IllegalArgumentException("cannot be greater than 65535!")));
    }
    switch (getProtocol()) {
      case LDAP: case LDAPS: case TLS:
        break;
      default:
        errors.add(new ConnectorError("Protocol",
            new IllegalArgumentException("unknown protocol")));
        break;
    }
    switch (getScope()) {
      case BASE: case ONELEVEL: case SUBTREE:
        break;
      default:
        errors.add(new ConnectorError("Scope",
            new IllegalArgumentException("unknown scope")));
      break;
    }
    if (getCertificatePath() != null) {
      File temp = new File(certificatePath);
      if (!temp.exists()) {
        errors.add(new ConnectorError("certificatePath",
            new IllegalArgumentException(
            "the given path does not refer to a real file!")));
      }
    }
    if (getSizeLimit() < 0) {
      errors.add(new ConnectorError("sizeLimit",
          new IllegalArgumentException("null or negative value is forbidden!")));
    }
    if (getTimeLimit() < 0) {
      errors.add(new ConnectorError("timeLimit",
          new IllegalArgumentException("null or negative value is forbidden!")));
    }
    if (!getReferralHandling().equals("ignore")
        && !getReferralHandling().equals("follow")) {
      errors.add(new ConnectorError("referralHandling",
          new IllegalArgumentException("must be either ignore or follow!")));
    }
    return errors;
  }
}
